<#
    .SYNOPSIS
    This script will install the provided AWSNVMe driver and, if requested,
    remove previous versions found.

    .PARAMETER NoReboot
    Do not reboot the instance automatically if necessary to complete driver
    install or uninstall.

    .PARAMETER RemovePreviousVersions
    Remove any other versions of the AWSNVMe driver found.
    *WARNING* Removing previous versions of the driver on Windows 2012 RTM
    (not R2) and then changing instance types to one with an NVMe boot volume
    may result in a bugcheck loop due to inaccesssable boot device.

    .EXAMPLE
    .\install.ps1

    .LEGAL
    Copyright (c) 2018 Amazon Web Services, Inc.  All rights reserved.

    .NOTES
    If you are installing this driver for the first time in order to migrate
    to a newer instance type with an NVMe boot volume it is important to go
    to the following link to ensure that the instance is bootable after
    changing the instance type and look for the section related to the NVMe
    drivers:
    https://docs.aws.amazon.com/AWSEC2/latest/WindowsGuide/migrating-latest-types.html

    If you have already applied those steps or are using a recent Amazon
    provided Windows AMI and just want to upgrade your drivers please make sure
    to note the following:

    If the script is run on 2012 RTM (2008R2, 2012R2, and 2016 work correctly)
    and the RemovePreviousVersions parameter is specified when running on an
    instance type that does not have an NVMe volume as it's boot device the
    instance will get stuck in a bugcheck loop due to an inaccessable boot
    device (0x7B) due to a problem in the way 2012 RTM processes driver
    installs for boot volume drivers. For 2012 RTM when an instance is migrated
    from a platform that does not have an NVMe boot disk to a platform that
    does on first boot the new driver will be loaded and used once kernel
    initialization has completed. However if the previous version of the driver
    isn't present it can't make it to that point during the initial boot on the
    new platform. Once the initial boot on the new instance type is complete it
    is safe to remove the old drivers. This limitation in 2012 RTM can be
    overcome by sysprepping the instance before migrating to the new instance
    type as documented in the link above regarding instance migration.

    .LINK
    https://docs.aws.amazon.com/AWSEC2/latest/WindowsGuide/aws-nvme-drivers.html
#>

Param (
    [parameter(Mandatory=$false)][switch]$NoReboot,
    [parameter(Mandatory=$false)][switch]$RemovePreviousVersions
)


# This function is basically a version of Get-FileHash that works on 2008R2+
# since that commandlet isn't available by default on it.
function Get-FileSHA256 {
    param (
        [parameter(Mandatory=$true)][string]$FilePath
    )
    $FilePath = Resolve-Path -Path $FilePath

    $algorithm = [Security.Cryptography.HashAlgorithm]::Create("SHA256")
    $fileStreamReader = [System.IO.StreamReader]$FilePath
    $hashBytes = $algorithm.ComputeHash($fileStreamReader.BaseStream)
    $hashString = (-Join ($hashBytes | ForEach-Object {"{0:X2}" -f $_}))
    $fileStreamReader.Close()

    return $hashString
}

$osVer = [System.Environment]::OSVersion.Version
if (($osVer.Major -lt 6) -or (($osVer.Major -eq 6) -and ($osVer.Minor -eq 0))) {
    Write-Error "Unsupported Version of Windows detected"
    return 1
}

if ($env:PROCESSOR_ARCHITECTURE -ne 'AMD64') {
    Write-Error "Unsupported Windows Platform detected"
    return 1
}

#Return codes for dpinst
$RETCODE_UNINSTALL_SUCCESSFUL                               = 0
$RETCODE_ONEPACKAGE_SUCCESSULLY_INSTALLED                   = 1
$RETCODE_TWOPACKAGES_SUCCESSFULLY_INSTALLED                 = 2
$RETCODE_INSTALLATION_SUCCESS_RESTART_REQUIRED              = 1073741825
$RETCODE_INSTALLATION_SUCCESS_2PACKAGES_RESTART_REQUIRED    = 1073741826
$RETCODE_DRIVER_STORE_COPIED_INSTALL_FAILED                 = 256
$RETCODE_INSTALLATION_FAILED                                = -2147549184
$RETCODE_ONEPACKAGE_INSTALLED_RESTARTREQUIRED_FOR_OTHER     = 3221291009
$RETCODE_NO_PACKAGE_FOUND                                   = -2147483648
$RETCODE_REBOOT_REQUIRED_TO_COMPLETE_UNINSTALL              = 1073741824

$exitCodeSuccessList = @(
    $RETCODE_UNINSTALL_SUCCESSFUL,
    $RETCODE_ONEPACKAGE_SUCCESSULLY_INSTALLED,
    $RETCODE_TWOPACKAGES_SUCCESSFULLY_INSTALLED
)

$exitCodeSuccessRestartList = @(
    $RETCODE_INSTALLATION_SUCCESS_RESTART_REQUIRED,
    $RETCODE_INSTALLATION_SUCCESS_2PACKAGES_RESTART_REQUIRED,
    $RETCODE_ONEPACKAGE_INSTALLED_RESTARTREQUIRED_FOR_OTHER,
    $RETCODE_REBOOT_REQUIRED_TO_COMPLETE_UNINSTALL
)

$exitCodeSuccessAllList = $exitCodeSuccessList + $exitCodeSuccessRestartList

$exitCodeFailedList = @(
    $RETCODE_DRIVER_STORE_COPIED_INSTALL_FAILED,
    $RETCODE_INSTALLATION_FAILED,
    $RETCODE_NO_PACKAGE_FOUND
)

# Need to support PowerShell 2 so we get script directory this way...
$scriptPath = Split-Path -Parent $MyInvocation.MyCommand.Definition
$infFile = Join-Path $scriptPath 'AWSNVMe.inf' -Resolve
$infFileHash = Get-FileSHA256 -FilePath $infFile
$dpinstFile = Join-Path $scriptPath 'dpinst.exe' -Resolve
$installArgList = @('/F', '/A', '/Q', '/SE', '/SA', '/PATH', $scriptPath)
$uninstallArgsListPartial = @('/Q', '/SE', '/F', '/D', '/U')

$rebootRequired = $false
$driverAlreadyInstalled = $false

$foundInfHashes = @()
# Get inf files for NVMe driver from file store and calculate hashes excluding
# the version we're installing
foreach ($f in (Get-ChildItem -Recurse -Path (Join-Path $env:SystemRoot 'system32\driverstore\filerepository' -Resolve) | Where-Object {$_.Name -eq 'AWSNVMe.inf'})) {
    $h = Get-FileSHA256 -FilePath $f.FullName
    if ($h -eq $infFileHash) {
        $driverAlreadyInstalled = $true
        Write-Output "Driver already installed in driver store under `"$($f.FullName)`"`nInstallation of this driver will be skipped."
    } else {
        $foundInfHashes += $h
    }
}

if ($driverAlreadyInstalled -eq $false) {
    Write-Output "Installing driver: `"$infFile`""
    $exitCode = (Start-Process -FilePath $dpinstFile -ArgumentList $installArgList -NoNewWindow -Wait -PassThru).ExitCode
    if (-not ($exitCode -in $exitCodeSuccessAllList)) {
        if ($exitCode -eq $RETCODE_DRIVER_STORE_COPIED_INSTALL_FAILED) {
            Write-Output 'WARNING: Device driver has been copied to the driver store however device'
            Write-Output 'installation was not complete. If you are running on an instance type without'
            Write-Output 'NVMe storage attached this error is not a problem. If you are running on an'
            Write-Output 'instance type with NVMe storage attached this may indicate a problem and'
            Write-Output 'should be manually investigated.'
            Write-Output ''
            Write-Output 'Execution continuing.'
        } else {
            Write-Error "Driver installation failed with exit code: $exitCode"
            return $exitCode
        }
    }
    if ($exitCode -in $exitCodeSuccessRestartList) {
        $rebootRequired = $true
    }
}

if ($RemovePreviousVersions) {
    $infsToRemove = @()
    foreach ($f in (Get-Item -Path (Join-Path $env:SystemRoot 'inf\oem*.inf'))) {
        $h = Get-FileSHA256 -FilePath $f.FullName
        if ($h -in $foundInfHashes) {
            $infsToRemove += $f
        }
    }

    foreach ($inf in $infsToRemove) {
        Write-Output "Removing version of the driver found here: $($inf.FullName)"
        if (($osVer.Major -eq 6) -and ($osVer.Minor -eq 2)) {
            Write-Error "WARNING: Windows 2012RTM detected. Removing the previous versions of the driver is not recommended as part of this scripted installation. See notes section of this script."
        }
        $exitCode = (Start-Process -FilePath $dpinstFile -ArgumentList ($uninstallArgsListPartial + $inf.FullName) -NoNewWindow -Wait -PassThru).ExitCode
        if (-not ($exitCode -in $exitCodeSuccessAllList)) {
            Write-Error "Driver removal failed with exit code: $exitCode"
            Write-Error "Continuing to process script"
        }
        if ($exitCode -in $exitCodeSuccessRestartList) {
            $rebootRequired = $true
        }
    }
}

Write-Output "Completed installation."

if ($rebootRequired) {
    if ($NoReboot -eq $false) {
        Write-Output "Starting reboot to complete driver install."
        Restart-Computer -Force
    } else {
        Write-Output "Please reboot the system to complete driver install/uninstall."
    }
}
